/* c_exceptions.c is part of CXCPT.
   CXCPT is a C exception-handling library.
   CXCPT is Copyright (C) 2002 Fergus Henderson.
   It may be copied under the terms of the GNU Library General Public License.

   This file implements the c_exceptions.h interface in C code, using
   setjmp() and longjmp().  It should be used only if compatibility
   with C++ exceptions is not important.  If compatibility with C++
   exceptions is desired, then c_exceptions.cpp should be used instead.

   Namespace cleanliness: all symbols exported by this module
   must start with the prefix `CXCPT_'. */

#include "c_exceptions.h"
#include <stdlib.h>
#include <setjmp.h>
#include <stdio.h>

struct CXCPT_ExceptionInfo {
	jmp_buf buf;
	const char *message;
	CXCPT_ExceptionKind kind;
	void *data;
};

struct CXCPT_ExceptionInfo *CXCPT_current;

void
CXCPT_throw(const char *message, CXCPT_ExceptionKind kind, void *data)
{
	if (CXCPT_current != NULL) {
		CXCPT_current->message = message;
		CXCPT_current->kind = kind;
		CXCPT_current->data = data;
		longjmp(CXCPT_current->buf, 1);
	} else {
		fprintf(stderr, "Uncaught exception: %s\n",
			(message != NULL ? message : "cause unknown"));
		exit(EXIT_FAILURE);
	}
}

void
CXCPT_try_catch(
	void (*action)(void *action_data),
	void *action_data,
	void (*handler)(void *handler_data, const char *message,
		CXCPT_ExceptionKind kind, void *exception_data),
	void *handler_data)
{
	struct CXCPT_ExceptionInfo e;
	struct CXCPT_ExceptionInfo *prev;

	prev = CXCPT_current;
	if (setjmp(e.buf) == 0) {
		CXCPT_current = &e;
		action(action_data);
		CXCPT_current = prev;
	} else {
		CXCPT_current = prev;
		handler(handler_data, e.message, e.kind, e.data);
	}
}

void
CXCPT_rethrow(void)
{
	longjmp(CXCPT_current->buf, 1);
}
