#include "c_exceptions.h"
#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static void
sort(void *base, size_t nmemb, size_t size, 
	int (*compar)(const void *, const void *))
{
	int num_sorted;
	int i, j;
	char *array = base;
	char *insert_elem;
	char *elem;

	/* insertion sort */
	for (num_sorted = 1; num_sorted < nmemb; num_sorted++) {
		insert_elem = &array[num_sorted * size];
		/* find where to put insert_elem */
		for (i = 0; (*compar)(insert_elem, &array[i * size]) > 0; i++)
			;
		elem = &array[i * size];
		/* swap elem and insert_elem */
		for (j = 0; j < size; j++) {
			char temp = elem[i];
			elem[i] = insert_elem[i];
			insert_elem[i] = temp;
		}
	}
}

static char compare_error[1];

static char *
to_lower(const char *line)
{
	size_t len = strlen(line);
	char *lower = malloc(len + 1);
	int i;
	for (i = 0; i < len + 1; i++) {
		lower[i] = tolower((unsigned char) line[i]);
	}
	return lower;
}

static int
do_compare(const void *data1, const void *data2) {
	char *const *line1 = data1;
	char *const *line2 = data2;
	char *lower_line1 = to_lower(*line1);
	char *lower_line2 = to_lower(*line2);
	int result = strcmp(lower_line1, lower_line2);

	if (result == 0 && strcmp(*line1, *line2) != 0) {
		CXCPT_throw("strings differ only in case",
			compare_error, NULL);
	}

	free(lower_line1);
	free(lower_line2);

	return result;
}

struct local_data {
	char **lines;
	int num_lines;
};

static void
do_sort(void *data) {
	struct local_data *local_data = data;
	sort(local_data->lines, local_data->num_lines,
		sizeof(local_data->lines[0]),
		do_compare);
}

static void
sort_handler(void *handler_data, const char *message,
	CXCPT_ExceptionKind kind, void *exception_data)
{
	if (kind == compare_error) {
		fprintf(stderr, "compare error: ");
		fprintf(stderr, "%s\n",
			(message != NULL ? message : "cause unknown"));
		exit(1);
	} else {
		CXCPT_rethrow();
	}
}

static void
sort_stdin(char **lines, int num_lines) {
	struct local_data data;
	data.lines = lines;
	data.num_lines = num_lines;
	CXCPT_try_catch(do_sort, &data, sort_handler, NULL);
}

static void
read_stdin(char **buf_ptr, char ***lines_ptr, int *num_lines_ptr)
{
	int c;
	int num_chars, num_lines;
	int char_num, line_num;
	char *buf;
	char **lines;

	/* compute the size of stdin */
	num_chars = num_lines = 0;
	while ((c = getchar()) != EOF) {
		num_chars++;
		if (c == '\n') {
			num_lines++;
		}
	}
	rewind(stdin);

	/* allocate a buffer of the appropriate size */
	buf = malloc(num_chars * sizeof(char));
	lines = malloc((num_lines + 1) * sizeof(char *));
	if (!buf || !lines) {
		fprintf(stderr, "Out of memory\n");
		exit(EXIT_FAILURE);
	}

	/* slurp stdin into the buffer */
	char_num = line_num = 0;
	lines[0] = buf;
	while ((c = getchar()) != EOF) {
		if (c == '\n') {
			buf[char_num++] = '\0';
			lines[++line_num] = &buf[char_num];
		} else {
			buf[char_num++] = c;
		}
	}
	assert(char_num == num_chars);
	assert(line_num == num_lines);

	*num_lines_ptr = num_lines;
	*lines_ptr = lines;
	*buf_ptr = buf;
}

static void
output_lines(char **lines, int num_lines)
{
	int i;
	for (i = 0; i < num_lines; i++) {
		puts(lines[i]);
	}
}

int
main(void)
{
	char *buf;
	char **lines;
	int num_lines;

	read_stdin(&buf, &lines, &num_lines);
	sort_stdin(lines, num_lines);
	output_lines(lines, num_lines);

	free(buf);
	free(lines);

	return 0;
}
